@@ -2,6 +2,7 @@
 
 import traceback
 
+from .utils import create_formatted_exception
 from .utils import format_kwargs
 from .utils import safe_native
 from .tokenize import Token
@@ -92,6 +93,20 @@ def compute_source_marker(line, column, expression, size):
     return s, column * " " + marker
 
 
+def iter_source_marker_lines(source, expression, line, column):
+    for i, l in enumerate(source):
+        if i + 1 != line:
+            continue
+
+        s, marker = compute_source_marker(
+            l, column, expression, LENGTH
+        )
+
+        yield " - Source:     %s" % s
+        yield "               %s" % marker
+        break
+
+
 def ellipsify(string, limit):
     if len(string) > limit:
         return "... " + string[-(limit - 4):]
@@ -99,10 +114,12 @@ def ellipsify(string, limit):
     return string
 
 
-def reconstruct_exc(cls, state):
-    exc = Exception.__new__(cls)
-    exc.__dict__ = state
-    return exc
+class RenderError(Exception):
+    """An error raised during rendering.
+
+    This class is used as a mixin which is added to the original
+    exception.
+    """
 
 
 class TemplateError(Exception):
@@ -121,7 +138,7 @@ class TemplateError(Exception):
     And pickle/unpickled:
 
     >>> from pickle import dumps, loads
-    >>> loads(dumps(TemplateError(message, token)))
+    >>> loads(dumps(TemplateError(message, token), -1))
     TemplateError('message', 'token')
 
     """
@@ -130,23 +147,16 @@ class TemplateError(Exception):
         if not isinstance(token, Token):
             token = Token(token, 0)
 
-        self.msg = msg
-        self.token = safe_native(token)
-        self.offset = getattr(token, "pos", 0)
-        self.filename = token.filename
-        self.location = token.location
+        Exception.__init__(self, msg, token)
 
     def __copy__(self):
         inst = Exception.__new__(type(self))
-        inst.__dict__ = self.__dict__.copy()
+        inst.args = self.args
         return inst
 
-    def __reduce__(self):
-        return reconstruct_exc, (type(self), self.__dict__)
-
     def __str__(self):
-        text = "%s\n\n" % self.msg
-        text += " - String:     \"%s\"" % self.token
+        text = "%s\n\n" % self.args[0]
+        text += " - String:     \"%s\"" % safe_native(self.token)
 
         if self.filename:
             text += "\n"
@@ -156,16 +166,58 @@ class TemplateError(Exception):
         text += "\n"
         text += " - Location:   (line %d: col %d)" % (line, column)
 
+        if line and column:
+            if self.token.source:
+                lines = iter_source_marker_lines(
+                    self.token.source.splitlines(),
+                    self.token, line, column
+                )
+            elif self.filename and not self.filename.startswith('<'):
+                try:
+                    f = open(self.filename, 'r')
+                except IOError:
+                    pass
+                else:
+                    it = iter_source_marker_lines(
+                        iter(f), self.token, line, column
+                    )
+                    try:
+                        lines = list(lines)
+                    finally:
+                        f.close()
+            else:
+                lines = ()
+
+            # Prepend newlines.
+            for line in lines:
+                text += "\n" + line
+
         return text
 
     def __repr__(self):
         try:
             return "%s('%s', '%s')" % (
-                self.__class__.__name__, self.msg, self.token
+                self.__class__.__name__, self.args[0], safe_native(self.token)
                 )
         except AttributeError:
             return object.__repr__(self)
 
+    @property
+    def token(self):
+        return self.args[1]
+
+    @property
+    def filename(self):
+        return self.token.filename
+
+    @property
+    def location(self):
+        return self.token.location
+
+    @property
+    def offset(self):
+        return getattr(self.token, "pos", 0)
+
 
 class ParseError(TemplateError):
     """An error occurred during parsing.
@@ -231,19 +283,12 @@ class ExceptionFormatter(object):
             formatted[index + 1] = " " * 15 + string
 
         out = []
-        seen = set()
 
-        for error in reversed(self._errors):
+        for error in self._errors:
             expression, line, column, filename, exc = error
 
-            if exc in seen:
-                continue
-
-            seen.add(exc)
-
             if isinstance(exc, UnicodeDecodeError):
                 string = safe_native(exc.object)
-
                 s, marker = compute_source_marker(
                     string, exc.start, string[exc.start:exc.end], LENGTH
                     )
@@ -257,28 +302,27 @@ class ExceptionFormatter(object):
             out.append(" - Filename:   %s" % _filename)
             out.append(" - Location:   (line %d: col %d)" % (line, column))
 
-            if filename and line and column:
+            if filename and not filename.startswith('<') and line and column:
                 try:
                     f = open(filename, 'r')
                 except IOError:
                     pass
                 else:
+                    lines = iter_source_marker_lines(
+                        iter(f), expression, line, column
+                    )
                     try:
-                        # Pick out source line and format marker
-                        for i, l in enumerate(f):
-                            if i + 1 == line:
-                                s, marker = compute_source_marker(
-                                    l, column, expression, LENGTH
-                                    )
-
-                                out.append(" - Source:     %s" % s)
-                                out.append("               %s" % marker)
-                                break
+                        out.extend(lines)
                     finally:
                         f.close()
 
         out.append(" - Arguments:  %s" % "\n".join(formatted))
 
+        if isinstance(exc.__str__, ExceptionFormatter):
+            # This is a nested error that has already been wrapped
+            # We must unwrap it before trying to format it to prevent
+            # recursion
+            exc = create_formatted_exception(exc, type(exc), exc._original__str__)
         formatted = traceback.format_exception_only(type(exc), exc)[-1]
         formatted_class = "%s:" % type(exc).__name__
 
