@@ -26,6 +26,7 @@ from ..loader import TemplateLoader
 from ..astutil import Builtin
 from ..utils import decode_string
 from ..utils import string_type
+from ..utils import unicode_string
 
 from .program import MacroProgram
 
@@ -44,6 +45,12 @@ class PageTemplate(BaseTemplate):
 
     Configuration (keyword arguments):
 
+      ``auto_reload``
+
+        Enables automatic reload of templates. This is mostly useful
+        in a development mode since it takes a significant performance
+        hit.
+
       ``default_expression``
 
         Set the default expression type. The default setting is
@@ -116,6 +123,11 @@ class PageTemplate(BaseTemplate):
 
           implicit_i18n_attributes = set(['alt', 'title'])
 
+      ``on_error_handler``
+
+        This is an optional exception handler that is invoked during the
+        "on-error" fallback block.
+
       ``strict``
 
         Enabled by default. If disabled, expressions are only required
@@ -129,7 +141,23 @@ class PageTemplate(BaseTemplate):
 
         If set, additional attribute whitespace will be stripped.
 
+      ``restricted_namespace``
+
+        True by default. If set False, ignored all namespace except chameleon default namespaces. It will be useful working with attributes based javascript template renderer like VueJS.
+
+        Example:
+
+          <div v-bind:id="dynamicId"></div>
+          <button v-on:click="greet">Greet</button>
+          <button @click="greet">Greet</button>
+
+      ``tokenizer``
+
+        None by default. If provided, this tokenizer is used instead of the default
+        (which is selected based on the template mode parameter.)
+
     Output is unicode on Python 2 and string on Python 3.
+
     """
 
     expression_types = {
@@ -161,6 +189,12 @@ class PageTemplate(BaseTemplate):
 
     enable_data_attributes = False
 
+    on_error_handler = None
+
+    restricted_namespace = True
+
+    tokenizer = None
+
     def __init__(self, body, **config):
         self.macros = Macros(self)
         super(PageTemplate, self).__init__(body, **config)
@@ -175,7 +209,7 @@ class PageTemplate(BaseTemplate):
     @property
     def engine(self):
         if self.literal_false:
-            default_marker = ast.Str(s="__default__")
+            default_marker = Builtin("__default")
         else:
             default_marker = Builtin("False")
 
@@ -191,7 +225,7 @@ class PageTemplate(BaseTemplate):
 
     def parse(self, body):
         if self.literal_false:
-            default_marker = ast.Str(s="__default__")
+            default_marker = Builtin("__default")
         else:
             default_marker = Builtin("False")
 
@@ -204,36 +238,45 @@ class PageTemplate(BaseTemplate):
             implicit_i18n_attributes=self.implicit_i18n_attributes,
             trim_attribute_space=self.trim_attribute_space,
             enable_data_attributes=self.enable_data_attributes,
-            )
+            restricted_namespace=self.restricted_namespace,
+            tokenizer=self.tokenizer
+        )
 
-    def render(self, encoding=None, translate=None, **vars):
+    def render(self, encoding=None, **_kw):
         """Render template to string.
 
-        The ``encoding`` and ``translate`` arguments are documented in
-        the template class constructor. If passed to this method, they
-        are used instead of the class defaults.
+        If providd, the ``encoding`` argument overrides the template
+        default value.
 
-        Additional arguments:
+        Additional keyword arguments are passed as template variables.
+
+        In addition, some also have a special meaning:
+
+          ``translate``
+
+            This keyword argument will override the default template
+            translate function.
 
           ``target_language``
 
-            This argument will be partially applied to the translation
-            function.
-
-            An alternative is thus to simply provide a custom
-            translation function which includes this information or
-            relies on a different mechanism.
+            This will be used as the default argument to the translate
+            function if no `i18n:target` value is provided.
 
+            If not provided, the `translate` function will need to
+            negotiate a language based on the provided context.
         """
 
-        non_trivial_translate = translate is not None
-        translate = translate if non_trivial_translate else self.translate or \
-                    type(self).translate
+        translate = _kw.get('translate')
+        if translate is not None:
+            has_translate = True
+        else:
+            has_translate = False
+            translate = self.translate
 
-        # Curry language parameter if non-trivial
-        target_language = vars.get('target_language')
-        if target_language is not None:
-            translate = partial(translate, target_language=target_language)
+            # This should not be necessary, but we include it for
+            # backward compatibility.
+            if translate is None:
+                translate = type(self).translate
 
         encoding = encoding if encoding is not None else self.encoding
         if encoding is not None:
@@ -247,18 +290,20 @@ class PageTemplate(BaseTemplate):
         else:
             decode = decode_string
 
-        setdefault = vars.setdefault
-        setdefault("__translate", translate)
-        setdefault("__convert", translate)
-        setdefault("__decode", decode)
+        target_language = _kw.get('target_language')
 
-        if non_trivial_translate:
-            vars['translate'] = translate
+        setdefault = _kw.setdefault
+        setdefault("__translate", translate)
+        setdefault("__convert",
+                   partial(translate, target_language=target_language))
+        setdefault("__decode", decode)
+        setdefault("target_language", None)
+        setdefault("__on_error_handler", self.on_error_handler)
 
         # Make sure we have a repeat dictionary
-        if 'repeat' not in vars: vars['repeat'] = RepeatDict({})
+        if 'repeat' not in _kw: _kw['repeat'] = RepeatDict({})
 
-        return super(PageTemplate, self).render(**vars)
+        return super(PageTemplate, self).render(**_kw)
 
     def include(self, *args, **kwargs):
         self.cook_check()
@@ -266,7 +311,9 @@ class PageTemplate(BaseTemplate):
 
     def digest(self, body, names):
         hex = super(PageTemplate, self).digest(body, names)
-        digest = md5(hex.encode('ascii'))
+        if isinstance(hex, unicode_string):
+            hex = hex.encode('utf-8')
+        digest = md5(hex)
         digest.update(';'.join(names).encode('utf-8'))
 
         for attr in (
